from unittest import TestCase
from unittest.mock import ANY, Mock, call, patch

import pandas as pd
from mlblocks import MLPipeline

from orion import benchmark
from orion.evaluation import CONTEXTUAL_METRICS as METRICS
from orion.evaluation import contextual_confusion_matrix


def test__sort_leaderboard_rank():
    rank = 'f1'
    metrics = METRICS
    score = pd.DataFrame({
        'pipeline': range(5),
        'f1': range(5),
    })

    expected_return = pd.DataFrame({
        'pipeline': range(5)[::-1],
        'rank': range(1, 6),
        'f1': range(5)[::-1],
    })

    returned = benchmark._sort_leaderboard(score, rank, metrics)
    pd.testing.assert_frame_equal(returned, expected_return)


def test__sort_leaderboard_rank_does_not_exist():
    rank = 'does not exist'
    metrics = {'f1': METRICS['f1']}
    score = pd.DataFrame({
        'pipeline': range(5),
        'f1': range(5),
    })

    expected_return = pd.DataFrame({
        'pipeline': range(5)[::-1],
        'rank': range(1, 6),
        'f1': range(5)[::-1],
    })

    returned = benchmark._sort_leaderboard(score, rank, metrics)
    pd.testing.assert_frame_equal(returned, expected_return)


def test__sort_leaderboard_no_rank():
    rank = None
    metrics = METRICS
    score = {k: range(5) for k in metrics.keys()}

    score['pipeline'] = range(5)
    score = pd.DataFrame(score)

    expected_return = score.iloc[::-1].reset_index(drop=True)
    expected_return['rank'] = range(1, 6)

    returned = benchmark._sort_leaderboard(score, rank, metrics)

    assert len(returned.columns) == len(expected_return.columns)
    assert sorted(returned.columns) == sorted(expected_return.columns)
    pd.testing.assert_frame_equal(returned, expected_return[returned.columns], check_dtype=False)


def test__detrend_signal_trend():
    df = pd.DataFrame({
        'timestamp': range(5),
        'value': range(5)
    })

    expected_return = pd.DataFrame({
        'timestamp': range(5),
        'value': [0.0] * 5,
    })

    returned = benchmark._detrend_signal(df, 'value')
    pd.testing.assert_frame_equal(returned, expected_return)


def test__detrend_signal_no_trend():
    df = pd.DataFrame({
        'timestamp': range(5),
        'value': [0.0] * 5
    })

    expected_return = df.copy()

    returned = benchmark._detrend_signal(df, 'value')
    pd.testing.assert_frame_equal(returned, expected_return)


def test__get_pipeline_hyperparameter():
    hyperparameters = {
        "pipeline1": "pipeline1.json",
        "pipeline2": "pipeline2.json",
    }
    pipeline = "pipeline1"

    expected_return = "pipeline1.json"
    returned = benchmark._get_pipeline_hyperparameter(hyperparameters, None, pipeline)
    assert returned == expected_return


def test__get_pipeline_hyperparameter_dataset():
    hyperparameters = {
        "dataset1": {
            "pipeline1": "pipeline1.json",
            "pipeline2": "pipeline2.json",
        }
    }
    dataset = "dataset1"

    expected_return = {
        "pipeline1": "pipeline1.json",
        "pipeline2": "pipeline2.json",
    }
    returned = benchmark._get_pipeline_hyperparameter(hyperparameters, dataset, None)
    assert returned == expected_return


def test__get_pipeline_hyperparameter_does_not_exist():
    hyperparameters = None
    pipeline = "pipeline1"

    expected_return = None
    returned = benchmark._get_pipeline_hyperparameter(hyperparameters, None, pipeline)
    assert returned == expected_return


@patch('orion.benchmark.load_signal')
def test__load_signal_test_split_true(load_signal_mock):
    train = Mock(autospec=pd.DataFrame)
    test = Mock(autospec=pd.DataFrame)

    load_signal_mock.return_value = (train, test)

    test_split = True
    returned = benchmark._load_signal('signal-name', test_split)

    assert isinstance(returned, tuple)
    assert len(returned) == 2

    expected_calls = [
        call('signal-name-train'),
        call('signal-name-test')
    ]
    assert load_signal_mock.call_args_list == expected_calls


@patch('orion.benchmark.load_signal')
def test__load_signal_test_split_false(load_signal_mock):
    df = pd.DataFrame({
        'timestamp': list(range(10)),
        'value': list(range(10, 20))
    })
    load_signal_mock.return_value = df

    test_split = False
    returned = benchmark._load_signal('signal-name', test_split)

    assert isinstance(returned, tuple)
    assert len(returned) == 2

    train, test = returned
    pd.testing.assert_frame_equal(train, test)

    expected_calls = [
        call('signal-name'),
    ]
    assert load_signal_mock.call_args_list == expected_calls


@patch('orion.benchmark.load_signal')
def test__load_signal_test_split_float(load_signal_mock):
    train = Mock(autospec=pd.DataFrame)
    test = Mock(autospec=pd.DataFrame)

    load_signal_mock.return_value = (train, test)

    test_split = 0.2
    returned = benchmark._load_signal('signal-name', test_split)

    assert isinstance(returned, tuple)
    assert len(returned) == 2

    expected_calls = [
        call('signal-name', test_size=test_split),
    ]
    assert load_signal_mock.call_args_list == expected_calls


class TestBenchmark(TestCase):

    @classmethod
    def setup_class(cls):
        cls.pipeline = Mock(autospec=MLPipeline)
        cls.name = 'pipeline-name'
        cls.dataset = 'dataset-name'
        cls.signal = 'signal-name'
        cls.hyper = None
        cls.distributed = False
        cls.rank = 'metric-name'
        cls.test_split = False
        cls.detrend = False
        cls.pipeline_path = None
        cls.cache_dir = None
        cls.iteration = 0
        cls.run_id = ANY
        cls.metrics = {
            'metric-name': Mock(autospec=METRICS['f1'], return_value=1)
        }

    def args(self):
        return (
            self.pipeline,
            self.name,
            self.dataset,
            self.signal,
            self.hyper,
            self.metrics,
            self.test_split,
            self.detrend,
            self.iteration,
            self.cache_dir,
            self.pipeline_path,
            self.run_id
        )

    def set_score(self, metric, elapsed, test_split):
        return {
            'metric-name': metric,
            'elapsed': elapsed,
            'split': test_split,
            'status': 'OK'
        }

    def set_output(self, metric, elapsed, test_split):
        return {
            'dataset': self.dataset,
            'pipeline': self.name,
            'signal': self.signal,
            'iteration': self.iteration,
            'metric-name': metric,
            'elapsed': elapsed,
            'split': test_split,
            'status': 'OK',
            'run_id': self.run_id
        }

    @patch('orion.benchmark.load_anomalies')
    @patch('orion.benchmark.analyze')
    @patch('orion.benchmark._load_pipeline')
    @patch('orion.benchmark.load_signal')
    def test__evaluate_signal(
            self, load_signal_mock, load_pipeline_mock, analyze_mock, load_anomalies_mock):
        train = Mock(autospec=pd.DataFrame)
        test = Mock(autospec=pd.DataFrame)
        load_signal_mock.side_effect = [train, test]
        load_pipeline_mock.return_value = self.pipeline

        anomalies = Mock(autospec=pd.DataFrame)
        analyze_mock.return_value = anomalies

        returned = benchmark._evaluate_signal(
            self.pipeline, self.signal, self.hyper, self.metrics, True)

        expected_return = self.set_score(1, ANY, ANY)
        assert returned == expected_return

        expected_calls = [
            call('signal-name-train'),
            call('signal-name-test')
        ]
        assert load_signal_mock.call_args_list == expected_calls

        load_pipeline_mock.assert_called_once_with(self.pipeline, self.hyper)
        analyze_mock.assert_called_once_with(self.pipeline, train, test)
        load_anomalies_mock.assert_called_once_with(self.signal)

    @patch('orion.benchmark.load_anomalies')
    @patch('orion.benchmark.analyze')
    @patch('orion.benchmark._load_pipeline')
    @patch('orion.benchmark.load_signal')
    def test__evaluate_signal_exception(
            self, load_signal_mock, load_pipeline_mock, analyze_mock, load_anomalies_mock):
        train = Mock(autospec=pd.DataFrame)
        test = Mock(autospec=pd.DataFrame)
        load_signal_mock.side_effect = [train, test]
        load_pipeline_mock.return_value = self.pipeline

        analyze_mock.side_effect = Exception("failed analyze.")

        returned = benchmark._evaluate_signal(
            self.pipeline, self.signal, self.hyper, self.metrics, True)

        expected_return = self.set_score(0, ANY, ANY)
        expected_return['status'] = 'ERROR'
        assert returned == expected_return

        expected_calls = [
            call('signal-name-train'),
            call('signal-name-test')
        ]
        assert load_signal_mock.call_args_list == expected_calls

        load_pipeline_mock.assert_called_once_with(self.pipeline, self.hyper)
        analyze_mock.assert_called_once_with(self.pipeline, train, test)

        assert load_anomalies_mock.called

    @patch('orion.benchmark.load_anomalies')
    @patch('orion.benchmark.analyze')
    @patch('orion.benchmark._load_pipeline')
    @patch('orion.benchmark.load_signal')
    def test__evaluate_signal_exception_confusion_matrix(
            self, load_signal_mock, load_pipeline_mock, analyze_mock, load_anomalies_mock):
        anomalies = pd.DataFrame({
            'start': [10, 35],
            'end': [20, 40]
        })

        train = Mock(autospec=pd.DataFrame)
        test = Mock(autospec=pd.DataFrame)
        load_signal_mock.side_effect = [train, test]
        load_pipeline_mock.return_value = self.pipeline
        load_anomalies_mock.return_value = anomalies
        analyze_mock.side_effect = Exception("failed analyze.")

        metrics = {'confusion_matrix': Mock(autospec=contextual_confusion_matrix)}
        metrics = {**metrics, **self.metrics}
        returned = benchmark._evaluate_signal(
            self.pipeline, self.signal, self.hyper, metrics, True)

        expected_return = self.set_score(0, ANY, ANY)
        expected_return['status'] = 'ERROR'
        expected_return['tn'] = None
        expected_return['fp'] = 0
        expected_return['fn'] = 2
        expected_return['tp'] = 0

        assert returned == expected_return

    @patch('orion.benchmark.load_anomalies')
    @patch('orion.benchmark.analyze')
    @patch('orion.benchmark._load_pipeline')
    @patch('orion.benchmark.load_signal')
    def test__evaluate_signal_test_split(
            self, load_signal_mock, load_pipeline_mock, analyze_mock, load_anomalies_mock):
        train = Mock(autospec=pd.DataFrame)
        test = Mock(autospec=pd.DataFrame)
        load_signal_mock.side_effect = [train, test]
        load_pipeline_mock.return_value = self.pipeline

        test_split = True

        returned = benchmark._evaluate_signal(
            self.pipeline, self.signal, self.hyper, self.metrics, test_split=test_split)

        expected_return = self.set_score(1, ANY, test_split)
        assert returned == expected_return

        expected_calls = [
            call('signal-name-train'),
            call('signal-name-test')
        ]
        assert load_signal_mock.call_args_list == expected_calls

        load_pipeline_mock.assert_called_once_with(self.pipeline, self.hyper)
        analyze_mock.assert_called_once_with(self.pipeline, train, test)
        load_anomalies_mock.assert_called_once_with(self.signal)

    @patch('orion.benchmark.load_anomalies')
    @patch('orion.benchmark.analyze')
    @patch('orion.benchmark._load_pipeline')
    @patch('orion.benchmark.load_signal')
    def test__evaluate_signal_no_test_split(
            self, load_signal_mock, load_pipeline_mock, analyze_mock, load_anomalies_mock):
        train = test = Mock(autospec=pd.DataFrame)
        load_signal_mock.side_effect = [train, test]
        load_pipeline_mock.return_value = self.pipeline

        test_split = False

        returned = benchmark._evaluate_signal(
            self.pipeline, self.signal, self.hyper, self.metrics, test_split=test_split)

        expected_return = self.set_score(1, ANY, test_split)
        assert returned == expected_return

        expected_calls = [
            call('signal-name')
        ]
        assert load_signal_mock.call_args_list == expected_calls

        load_pipeline_mock.assert_called_once_with(self.pipeline, self.hyper)
        analyze_mock.assert_called_once_with(self.pipeline, train, test)
        load_anomalies_mock.assert_called_once_with(self.signal)

    @patch('orion.benchmark.load_anomalies')
    @patch('orion.benchmark.analyze')
    @patch('orion.benchmark._load_pipeline')
    @patch('orion.benchmark.load_signal')
    def test__evaluate_signal_no_detrend(
            self, load_signal_mock, load_pipeline_mock, analyze_mock, load_anomalies_mock):
        train = Mock(autospec=pd.DataFrame)
        test = Mock(autospec=pd.DataFrame)
        load_signal_mock.side_effect = [train, test]
        load_pipeline_mock.return_value = self.pipeline

        detrend = False

        returned = benchmark._evaluate_signal(
            self.pipeline, self.signal, self.hyper, self.metrics, test_split=True, detrend=detrend)

        expected_return = self.set_score(1, ANY, ANY)
        assert returned == expected_return

        expected_calls = [
            call('signal-name-train'),
            call('signal-name-test')
        ]
        assert load_signal_mock.call_args_list == expected_calls

        load_pipeline_mock.assert_called_once_with(self.pipeline, self.hyper)
        analyze_mock.assert_called_once_with(self.pipeline, train, test)
        load_anomalies_mock.assert_called_once_with(self.signal)

    @patch('orion.benchmark.load_anomalies')
    @patch('orion.benchmark.analyze')
    @patch('orion.benchmark._load_pipeline')
    @patch('orion.benchmark.load_signal')
    @patch('orion.benchmark._detrend_signal')
    def test__evaluate_signal_detrend(self, detrend_signal_mock, load_signal_mock,
                                      load_pipeline_mock, analyze_mock, load_anomalies_mock):

        train = Mock(autospec=pd.DataFrame)
        test = Mock(autospec=pd.DataFrame)
        detrend_signal_mock.side_effect = [train, test]
        load_signal_mock.side_effect = [train, test]
        load_pipeline_mock.return_value = self.pipeline

        detrend = True

        returned = benchmark._evaluate_signal(
            self.pipeline, self.signal, self.hyper, self.metrics, test_split=True, detrend=detrend)

        expected_return = self.set_score(1, ANY, ANY)
        assert returned == expected_return

        expected_calls = [
            call('signal-name-train'),
            call('signal-name-test')
        ]
        assert load_signal_mock.call_args_list == expected_calls

        expected_calls = [
            call(train, 'value'),
            call(test, 'value')
        ]
        assert detrend_signal_mock.call_args_list == expected_calls

        load_pipeline_mock.assert_called_once_with(self.pipeline, self.hyper)
        analyze_mock.assert_called_once_with(self.pipeline, train, test)
        load_anomalies_mock.assert_called_once_with(self.signal)

    @patch('orion.benchmark._evaluate_signal')
    def test__run_job(self, evaluate_signal_mock):
        pass

        score = self.set_score(1, ANY, ANY)
        evaluate_signal_mock.return_value = score

        benchmark._run_job(self.args())

        expected_calls = [
            self.pipeline, self.signal, self.hyper, self.metrics,
            self.test_split, self.detrend, self.pipeline_path
        ]

        evaluate_signal_mock.assert_called_once_with(*expected_calls)

    @patch('orion.benchmark._run_job')
    def test_benchmark(self, run_job_mock):
        signals = [self.signal]
        datasets = {self.dataset: signals}
        pipelines = {self.name: self.pipeline}

        output = self.set_output(1, ANY, ANY)
        run_job_mock.return_value = pd.DataFrame.from_records([output])

        order = [
            'pipeline',
            'rank',
            'dataset',
            'signal',
            'iteration',
            'metric-name',
            'status',
            'elapsed',
            'split',
            'run_id']

        expected_return = pd.DataFrame.from_records([{
            'metric-name': 1,
            'rank': 1,
            'elapsed': ANY,
            'split': ANY,
            'status': 'OK',
            'iteration': self.iteration,
            'pipeline': self.name,
            'dataset': self.dataset,
            'signal': self.signal,
            'run_id': self.run_id
        }])[order]

        returned = benchmark.benchmark(
            pipelines, datasets, self.hyper, self.metrics, self.rank)[order]

        pd.testing.assert_frame_equal(returned, expected_return)

        run_job_mock.assert_called_once_with(self.args())

    @patch('orion.benchmark._run_job')
    def test_benchmark_dataset_list(self, run_job_mock):
        signals = [self.signal]
        pipelines = {self.name: self.pipeline}

        output = self.set_output(1, ANY, ANY)
        output['dataset'] = 'dataset'
        run_job_mock.return_value = pd.DataFrame.from_records([output])

        order = [
            'pipeline',
            'rank',
            'dataset',
            'signal',
            'iteration',
            'metric-name',
            'status',
            'elapsed',
            'split',
            'run_id']

        expected_return = pd.DataFrame.from_records([{
            'metric-name': 1,
            'rank': 1,
            'elapsed': ANY,
            'split': ANY,
            'status': 'OK',
            'iteration': self.iteration,
            'pipeline': self.name,
            'dataset': 'dataset',
            'signal': self.signal,
            'run_id': self.run_id
        }])[order]

        returned = benchmark.benchmark(
            pipelines, signals, self.hyper, self.metrics, self.rank)[order]

        pd.testing.assert_frame_equal(returned, expected_return)

        args = list(self.args())
        args[2] = 'dataset'
        run_job_mock.assert_called_once_with(tuple(args))

    @patch('orion.benchmark._run_job')
    def test_benchmark_pipelines_list(self, run_job_mock):
        signals = [self.signal]
        datasets = {self.dataset: signals}
        pipelines = [self.pipeline]

        hyper = {}

        output = self.set_output(1, ANY, ANY)
        output['pipeline'] = self.pipeline
        run_job_mock.return_value = pd.DataFrame.from_records([output])

        order = [
            'pipeline',
            'rank',
            'dataset',
            'signal',
            'iteration',
            'metric-name',
            'status',
            'elapsed',
            'split',
            'run_id']

        expected_return = pd.DataFrame.from_records([{
            'metric-name': 1,
            'rank': 1,
            'elapsed': ANY,
            'split': ANY,
            'status': 'OK',
            'iteration': self.iteration,
            'pipeline': self.pipeline,
            'dataset': self.dataset,
            'signal': self.signal,
            'run_id': self.run_id
        }])[order]

        returned = benchmark.benchmark(
            pipelines, datasets, hyper, self.metrics, self.rank)[order]

        pd.testing.assert_frame_equal(returned, expected_return)

        args = list(self.args())
        args[1] = self.pipeline
        args[4] = hyper

        run_job_mock.assert_called_once_with(tuple(args))

    @patch('orion.benchmark._run_job')
    def test_benchmark_metrics_list(self, run_job_mock):
        pass

        signals = [self.signal]
        datasets = {self.dataset: signals}
        pipelines = {self.name: self.pipeline}

        metric = Mock(autospec=METRICS['f1'], return_value=1)
        metric.__name__ = 'metric-name'
        metrics = [metric]
        metrics_ = {metric.__name__: metric}

        output = self.set_output(1, ANY, ANY)
        output[metric.__name__] = metric
        run_job_mock.return_value = pd.DataFrame.from_records([output])

        order = [
            'pipeline',
            'rank',
            'dataset',
            'signal',
            'iteration',
            'metric-name',
            'status',
            'elapsed',
            'split',
            'run_id']

        expected_return = pd.DataFrame.from_records([{
            'metric-name': metric,
            'rank': 1,
            'elapsed': ANY,
            'split': ANY,
            'status': 'OK',
            'iteration': self.iteration,
            'pipeline': self.name,
            'dataset': self.dataset,
            'signal': self.signal,
            'run_id': self.run_id
        }])[order]

        returned = benchmark.benchmark(
            pipelines, datasets, self.hyper, metrics, self.rank)[order]

        pd.testing.assert_frame_equal(returned, expected_return)

        args = list(self.args())
        args[5] = metrics_

        run_job_mock.assert_called_once_with(tuple(args))

    def test_benchmark_metrics_exception(self):
        pass

        signals = [self.signal]
        datasets = {self.dataset: signals}
        pipelines = {self.name: self.pipeline}

        metric = 'does-not-exist'
        metrics = [metric]

        with self.assertRaises(ValueError) as ex:
            benchmark.benchmark(pipelines, datasets, self.hyper, metrics, self.rank)

            self.assertTrue(metric in ex.exception)



# -*- coding: utf-8 -*-

"""
Data Management module.

This module contains functions that allow downloading demo data from Amazon S3,
as well as load and work with other data stored locally.

The demo data is a modified version of the NASA data found here:

https://s3-us-west-2.amazonaws.com/telemanom/data.zip
"""

import json
import logging
import os

import numpy as np
import pandas as pd

LOGGER = logging.getLogger(__name__)

DATA_PATH = os.path.join(
    os.path.dirname(os.path.abspath(__file__)),
    'data'
)
BUCKET = 'd3-ai-orion'
S3_URL = 'https://{}.s3.amazonaws.com/{}'

NASA_SIGNALS = (
    'P-1', 'S-1', 'E-1', 'E-2', 'E-3', 'E-4', 'E-5', 'E-6', 'E-7',
    'E-8', 'E-9', 'E-10', 'E-11', 'E-12', 'E-13', 'A-1', 'D-1', 'P-3',
    'D-2', 'D-3', 'D-4', 'A-2', 'A-3', 'A-4', 'G-1', 'G-2', 'D-5',
    'D-6', 'D-7', 'F-1', 'P-4', 'G-3', 'T-1', 'T-2', 'D-8', 'D-9',
    'F-2', 'G-4', 'T-3', 'D-11', 'D-12', 'B-1', 'G-6', 'G-7', 'P-7',
    'R-1', 'A-5', 'A-6', 'A-7', 'D-13', 'A-8', 'A-9', 'F-3', 'M-6',
    'M-1', 'M-2', 'S-2', 'P-10', 'T-4', 'T-5', 'F-7', 'M-3', 'M-4',
    'M-5', 'P-15', 'C-1', 'C-2', 'T-12', 'T-13', 'F-4', 'F-5', 'D-14',
    'T-9', 'P-14', 'T-8', 'P-11', 'D-15', 'D-16', 'M-7', 'F-8'
)


def download(name, test_size=None, data_path=DATA_PATH):
    """Load the CSV with the given name from S3.

    If the CSV has never been loaded before, it will be downloaded
    from the [d3-ai-orion bucket](https://d3-ai-orion.s3.amazonaws.com) or
    the S3 bucket specified following the `s3://{bucket}/path/to/the.csv` format,
    and then cached inside the `data` folder, within the `orion` package
    directory, and then returned.

    Otherwise, if it has been downloaded and cached before, it will be directly
    loaded from the `orion/data` folder without contacting S3.

    If a `test_size` value is given, the data will be split in two parts
    without altering its order, making the second one proportionally as
    big as the given value.

    Args:
        name (str): Name of the CSV to load.
        test_size (float): Value between 0 and 1 indicating the proportional
            size of the test split. If 0 or None (default), the data is not split.

    Returns:
        If no test_size is given, a single pandas.DataFrame is returned containing all
        the data. If test_size is given, a tuple containing one pandas.DataFrame for
        the train split and another one for the test split is returned.
    """

    url = None
    if name.startswith('s3://'):
        parts = name[5:].split('/', 1)
        bucket = parts[0]
        path = parts[1]
        url = S3_URL.format(bucket, path)

        filename = os.path.join(data_path, path.split('/')[-1])
    else:
        filename = os.path.join(data_path, name + '.csv')
        data_path = os.path.join(data_path, os.path.dirname(name))

    if os.path.exists(filename):
        data = pd.read_csv(filename)
    else:
        url = url or S3_URL.format(BUCKET, '{}.csv'.format(name))

        LOGGER.info('Downloading CSV %s from %s', name, url)
        os.makedirs(data_path, exist_ok=True)
        data = pd.read_csv(url)
        data.to_csv(filename, index=False)

    return data


def download_demo(path='orion-data', split=False):
    if not os.path.exists(path):
        os.makedirs(path, exist_ok=True)

    LOGGER.info('Downloading Orion Demo Data to folder %s', path)
    for signal in NASA_SIGNALS[0:3]:
        if split:
            download(signal + '-train', data_path=path)
            download(signal + '-test', data_path=path)
        else:
            download(signal, data_path=path)


def format_csv(df, timestamp_column=None, value_columns=None):
    timestamp_column_name = df.columns[timestamp_column] if timestamp_column else df.columns[0]
    value_column_names = df.columns[value_columns] if value_columns else df.columns[1:]

    data = dict()
    data['timestamp'] = df[timestamp_column_name].astype('int64').values
    for column in value_column_names:
        data[column] = df[column].astype(float).values

    return pd.DataFrame(data)


def load_csv(path, timestamp_column=None, value_column=None):
    header = None if timestamp_column is not None else 'infer'
    data = pd.read_csv(path, header=header)

    if timestamp_column is None:
        if value_column is not None:
            raise ValueError("If value_column is provided, timestamp_column must be as well")

        return data

    elif value_column is None:
        raise ValueError("If timestamp_column is provided, value_column must be as well")
    elif timestamp_column == value_column:
        raise ValueError("timestamp_column cannot be the same as value_column")

    return format_csv(data, timestamp_column, value_column)


def load_signal(signal, test_size=None, timestamp_column=None, value_column=None):
    if os.path.isfile(signal):
        data = load_csv(signal, timestamp_column, value_column)
    else:
        data = download(signal)

    data = format_csv(data)

    if test_size is None:
        return data

    test_length = round(len(data) * test_size)
    train = data.iloc[:-test_length]
    test = data.iloc[-test_length:]

    return train, test


def load_anomalies(signal, edges=False):
    anomalies = download('anomalies')

    anomalies = anomalies.set_index('signal').loc[signal].values[0]
    anomalies = pd.DataFrame(json.loads(anomalies), columns=['start', 'end'])

    if edges:
        data = download(signal)
        start = data.timestamp.min()
        end = data.timestamp.max()

        anomalies['score'] = 1
        parts = np.concatenate([
            [[start, anomalies.start.min(), 0]],
            anomalies.values,
            [[anomalies.end.max(), end, 0]]
        ], axis=0)
        anomalies = pd.DataFrame(parts, columns=['start', 'end', 'score'])

    return anomalies



#!/usr/bin/env python
# coding: utf-8

import argparse
import logging
import warnings

import tabulate

from orion.analysis import get_available_templates
from orion.benchmark import benchmark
from orion.evaluation import CONTEXTUAL_METRICS as METRICS

warnings.filterwarnings("ignore")


def _evaluate(args):
    if args.all:
        pipelines = get_available_templates()
    else:
        pipelines = args.pipeline

    scores = benchmark(pipelines=pipelines, datasets=args.signal, metrics=args.metric,
                       rank=args.rank, test_split=args.holdout)

    if args.output:
        print('Writing results in {}'.format(args.output))
        scores.to_csv(args.output, index=False)

    print(tabulate.tabulate(
        scores,
        showindex=False,
        tablefmt='github',
        headers=scores.columns
    ))


def get_parser():

    # Common Parent - Shared options
    base = argparse.ArgumentParser(add_help=False)
    base.add_argument('-l', '--logfile',
                      help='Name of the logfile. If not given, log to stdout.')
    base.add_argument('-v', '--verbose', action='count', default=0,
                      help='Be verbose. Use -vv for increased verbosity.')

    parser = argparse.ArgumentParser(description='Orion Command Line Interface.')
    parser.set_defaults(function=None)

    action = parser.add_subparsers(title='action', dest='action')
    action.required = True

    # Evaluate
    evaluate = action.add_parser('evaluate', parents=[base],
                                 help='Evaluate one or more pipelines on NASA signals')
    evaluate.add_argument('-s', '--signal', action='append',
                          help='Signal to use. Use multiple times for more signals.')
    evaluate.add_argument('-m', '--metric', action='append', default=METRICS,
                          help='Metric to use. Use multiple times for more metrics.')
    evaluate.add_argument('-r', '--rank', default='f1', help='Rank scores based on this metric.')
    evaluate.add_argument('-o', '--output', help='Write the results in the specified CSV file.')
    evaluate.add_argument('--holdout', dest='holdout', action='store_true', default=None,
                          help='Holdout test data during training.')
    evaluate.add_argument('--no-holdout', dest='holdout', action='store_false', default=None,
                          help='Do not holdout test data curing training.')
    evaluate.set_defaults(function=_evaluate)

    group = evaluate.add_mutually_exclusive_group(required=True)
    group.add_argument('-p', '--pipeline', default=[], action='append',
                       help='Name of the pipeline JSONs to evaluate.')
    group.add_argument('-a', '--all', action='store_true', help='Evaluate all known pipelines.')

    return parser


def logging_setup(verbosity=1, logfile=None, logger_name=None):
    logger = logging.getLogger(logger_name)
    log_level = (3 - verbosity) * 10
    fmt = '%(asctime)s - %(process)d - %(levelname)s - %(module)s - %(message)s'
    formatter = logging.Formatter(fmt)
    logger.setLevel(log_level)
    logger.propagate = False

    if logfile:
        file_handler = logging.FileHandler(logfile)
        file_handler.setLevel(logging.DEBUG)
        file_handler.setFormatter(formatter)
        logger.addHandler(file_handler)

    else:
        console_handler = logging.StreamHandler()
        console_handler.setLevel(log_level)
        console_handler.setFormatter(formatter)
        logger.addHandler(console_handler)


def main():
    parser = get_parser()
    args = parser.parse_args()

    logging_setup(args.verbose, args.logfile)

    if args.action == 'evaluate':
        args.function(args)


if __name__ == "__main__":
    main()



import numpy as np
import pandas as pd

# helper functions


def from_pandas_contextual(df):
    """ Convert contextual ``pandas.DataFrame`` to list of tuples.

    Args:
        df (DataFrame):
            anomalies, passed as ``pandas.DataFrame``
            containing two columns: start and stop.

    Returns:
        list:
            tuple (start, end) timestamp.

    Raises:
        KeyError:
            If the received ``pandas.DataFrame`` does not contain the required columns.
    """
    require = ['start', 'end']
    columns = df.columns.tolist()

    if all(x in columns for x in require):
        if 'severity' in columns:
            return list(df[require + ['severity']].itertuples(index=False))
        return list(df[require].itertuples(index=False))

    raise KeyError('{} not found in columns: {}.'.format(require, columns))


def from_list_points_timestamps(timestamps, gap=1):
    """ Convert list of timestamps to list of tuples.

    Convert a list of anomalies identified by timestamps,
    to a list of tuples marking the start and end interval
    of anomalies; make it contextually defined.

    Args:
        timestamps (list): contains timestamp of anomalies.
        gap (int): allowed gap between anomalies.

    Returns:
        list:
            tuple (start, end) timestamp.
    """
    timestamps = sorted(timestamps)

    start_ts = 0
    max_ts = len(timestamps) - 1

    anomalies = list()
    break_point = start_ts
    while break_point < max_ts:
        if timestamps[break_point + 1] - timestamps[break_point] <= gap:
            break_point += 1
            continue

        anomalies.append((timestamps[start_ts], timestamps[break_point]))
        break_point += 1
        start_ts = break_point

    anomalies.append((timestamps[start_ts], timestamps[break_point]))
    return anomalies


def from_pandas_points(df):
    """ Convert point ``pandas.DataFrame`` to list of tuples.

    Convert a ``pandas.DataFrame`` of anomalies identified by
    one column (timestamp) to a list of tuples marking the
    start and end interval of anomalies; make it contextually
    defined.

    Args:
        df (DataFrame):
            anomalies, passed as ``pandas.DataFrame``
            containing one column: timestamp.

    Returns:
        list:
            tuple (start, end) timestamp.

    Raises:
        KeyError:
            If the received ``pandas.DataFrame`` does not contain column `timestamp`.
    """
    time_column = 'timestamp'
    columns = df.columns.tolist()
    if time_column not in columns:
        raise KeyError('{} not found in columns: {}.'.format(time_column, columns))

    timestamps = list(df['timestamp'])

    return from_list_points_timestamps(timestamps)


def from_pandas_points_labels(df):
    """ Convert point ``pandas.DataFrame`` to list of tuples.

    Convert a ``pandas.DataFrame`` of labeled data where each
    timestamp is labeled by either 0 or 1 to a list of tuples
    marking the start and end interval of anomalies; make it
    contextually defined.

    Args:
        df (DataFrame):
            anomalies, passed as ``pandas.DataFrame``
            containing two columns: timestamp and label.

    Returns:
        list:
            tuple (start, end) timestamp.

    Raises:
        KeyError:
            If the received ``pandas.DataFrame`` does not contain the required columns.
    """

    require = ['timestamp', 'label']
    columns = df.columns.tolist()
    if not all(x in columns for x in require):
        raise KeyError('{} not found in columns: {}.'.format(require, columns))

    df = df[df['label'] == 1]
    return from_pandas_points(df)


def from_list_points_labels(labels):
    """ Convert list of labels to list of tuples.

    Convert a list of labels to a list of tuples
    marking the start and end interval of anomalies by
    defining a dummy timestamp range for usage.

    Args:
        labels (list): contains binary labels [0, 1].

    Returns:
        list:
            tuple (start, end) timestamp.
    """

    timestamps = np.arange(len(labels))
    return from_pandas_points_labels(pd.DataFrame({"timestamp": timestamps, "label": labels}))



import numpy as np
import pandas as pd
import pytest
from numpy.testing import assert_array_equal

from orion.primitives.timeseries_preprocessing import fillna, slice_array_by_dims


def signal():
    return np.array([0, 3, np.nan, 4, 3, 2, 2, np.nan, np.nan]).reshape(-1, 1)


def multidimensional():
    return np.array([[0, 3, np.nan],
                     [np.nan, 2, 4],
                     [5, np.nan, 6],
                     [9, np.nan, 3],
                     [4, 7, np.nan]])


def series():
    X = signal()
    return pd.Series(X.flatten())


def test_slice_dim_axis_one():
    X = multidimensional()
    expected_return = X[:, [0]]
    returned = slice_array_by_dims(X, 0, axis=1)

    assert_array_equal(returned, expected_return)


def test_slice_dim_axis_zero():
    X = multidimensional()
    expected_return = X[[0], :]
    returned = slice_array_by_dims(X, 0, axis=0)

    assert_array_equal(returned, expected_return)


def test_slice_dim_identity():
    X = signal()
    returned = slice_array_by_dims(X, 0, axis=1)

    assert_array_equal(returned, X)


def test_slice_dim_error():
    X = multidimensional()
    with pytest.raises(ValueError):
        slice_array_by_dims(X, 0, axis=3)


def test_fillna_series():
    df = series()
    expected_return = np.array([0, 3, 0, 4, 3, 2, 2, 0, 0])

    returned = fillna(df, value=0)

    assert_array_equal(returned, expected_return)


def test_fillna_single_dimension():
    X = signal().flatten()
    expected_return = np.array([0, 3, 0, 4, 3, 2, 2, 0, 0])

    returned = fillna(X, value=0)

    assert_array_equal(returned, expected_return)


def test_fillna_ffill():
    X = signal()
    expected_return = np.array([0, 3, 3, 4, 3, 2, 2, 2, 2]).reshape(-1, 1)

    returned = fillna(X, method='ffill')

    assert_array_equal(returned, expected_return)


def test_fillna_bfill():
    X = signal()
    expected_return = np.array([0, 3, 4, 4, 3, 2, 2, np.nan, np.nan]).reshape(-1, 1)

    returned = fillna(X, method='bfill')

    assert_array_equal(returned, expected_return)


def test_fillna_ffill_bfill():
    X = signal()
    expected_return = np.array([0, 3, 3, 4, 3, 2, 2, 2, 2]).reshape(-1, 1)

    returned = fillna(X, method=['ffill', 'bfill'])

    assert_array_equal(returned, expected_return)


def test_fillna_multi_dimensional():
    X = multidimensional()
    expected_return = np.array([[0, 3, 4],
                                [0, 2, 4],
                                [5, 2, 6],
                                [9, 2, 3],
                                [4, 7, 3]])

    returned = fillna(X, method=['ffill', 'bfill'])

    assert_array_equal(returned, expected_return)



import logging

import numpy as np
from sklearn import metrics

LOGGER = logging.getLogger(__name__)


def _overlap(expected, observed):
    first = expected[0] - observed[1]
    second = expected[1] - observed[0]
    return first * second < 0


def _any_overlap(part, intervals):
    for interval in intervals:
        if _overlap(part, interval):
            return 1

    return 0


def _weighted_segment(expected, observed, _partition, start=None, end=None):
    expected, observed, weights = _partition(expected, observed, start, end)

    return metrics.confusion_matrix(
        expected, observed, sample_weight=weights, labels=[0, 1]).ravel()


def _accuracy(expected, observed, data, start, end, cm):
    tn, fp, fn, tp = cm(expected, observed, data, start, end)

    if tn is None:
        raise ValueError("Cannot obtain accuracy score for overlap segment method.")

    return (tp + tn) / (tn + fp + fn + tp)


def _precision(expected, observed, data, start, end, cm):
    tn, fp, fn, tp = cm(expected, observed, data, start, end)

    try:
        return tp / (tp + fp)

    except ZeroDivisionError as ex:
        LOGGER.exception(
            'Evaluation exception {} (tp {}/ fp {}).'.format(ex, tp, fp))

        return np.nan


def _recall(expected, observed, data, start, end, cm):
    tn, fp, fn, tp = cm(expected, observed, data, start, end)

    try:
        return tp / (tp + fn)

    except ZeroDivisionError as ex:
        LOGGER.exception(
            'Evaluation exception {} (tp {}/ fn {}).'.format(ex, tp, fn))

        return np.nan


def _f1_score(expected, observed, data, start, end, cm):
    precision = _precision(expected, observed, data, start, end, cm)
    recall = _recall(expected, observed, data, start, end, cm)

    try:
        return 2 * (precision * recall) / (precision + recall)

    except ZeroDivisionError:
        LOGGER.exception(
            'Invalid value encountered for precision {}/ recall {}.'.format(precision, recall))

        return np.nan



import numpy as np


def build_anomaly_intervals(X, y, time_column, severity=True, indices=False):
    """Group together consecutive anomalous samples in anomaly intervals.

    This is a dummy boundary detection function that groups together
    samples that have been consecutively flagged as anomalous and
    returns boundaries of anomalous intervals.

    Optionally, it computes the average severity of each interval.

    This detector is here only to serve as reference of what
    an boundary detection primitive looks like, and is not intended
    to be used in real scenarios.
    """

    timestamps = X[time_column]

    start = None
    start_ts = None
    intervals = list()
    values = list()
    for index, (value, timestamp) in enumerate(zip(y, timestamps)):
        if value != 0:
            if start_ts is None:
                start = index
                start_ts = timestamp
            if severity:
                values.append(value)

        elif start_ts is not None:
            interval = [start_ts, timestamp]
            if indices:
                interval.extend([start, index])
            if severity:
                interval.append(np.mean(values))
                values = list()

            intervals.append(tuple(interval))

            start = None
            start_ts = None

    # We might have an open interval at the end
    if start_ts is not None:
        interval = [start_ts, timestamp]
        if indices:
            interval.extend([start, index])
        if severity:
            interval.append(np.mean(values))

        intervals.append(tuple(interval))

    return np.array(intervals)



import numpy as np
import pandas as pd
import pytest

from orion.evaluation.contextual import (
    _contextual_partition, _overlap_segment, contextual_accuracy, contextual_confusion_matrix,
    contextual_f1_score, contextual_precision, contextual_recall)


@pytest.fixture()
def expected():
    return pd.DataFrame({'start': [1, 5, 7, 12, 16, 22], 'end': [2, 6, 10, 13, 17, 23]})


@pytest.fixture()
def observed():
    return pd.DataFrame({'start': [2, 4, 8, 11, 19, 22], 'end': [3, 5, 9, 14, 20, 23]})


@pytest.fixture()
def expected_point():
    return pd.DataFrame({'start': [2, 5, 8, 12, 15, 18], 'end': [2, 5, 8, 12, 17, 18]})


@pytest.fixture()
def observed_point():
    return pd.DataFrame({'start': [2, 3, 8, 11, 16, 20], 'end': [2, 5, 10, 13, 16, 20]})


def test__contextual_partition(expected, observed):
    expected = list(expected[['start', 'end']].itertuples(index=False))
    observed = list(observed[['start', 'end']].itertuples(index=False))
    expected_parts = [1, 0, 0, 0, 1, 0, 1, 1, 1, 0, 0, 1, 0, 0, 1, 0, 0, 0, 1]
    observed_parts = [0, 1, 0, 1, 0, 0, 0, 1, 0, 0, 1, 1, 1, 0, 0, 0, 1, 0, 1]
    weights = [1] * len(observed_parts)
    weights[13] = weights[15] = weights[17] = 2
    length = expected[-1][-1] - expected[0][0]

    expected_return, observed_return, weights_return = _contextual_partition(expected, observed)

    np.testing.assert_array_equal(np.array(expected_return),
                                  np.array(expected_parts))

    np.testing.assert_array_equal(np.array(observed_return),
                                  np.array(observed_parts))

    np.testing.assert_array_equal(np.array(weights_return),
                                  np.array(weights))

    assert len(expected_return) == len(observed_return) == len(weights_return)
    assert sum(weights_return) == length


def test__overlap_segment(expected, observed):
    expected = list(expected[['start', 'end']].itertuples(index=False))
    observed = list(observed[['start', 'end']].itertuples(index=False))
    expected_return = (None, 3, 3, 3)
    returned = _overlap_segment(expected, observed)
    np.testing.assert_array_equal(np.array(returned),
                                  np.array(expected_return))


def test__overlap_segment_points(expected_point, observed_point):
    expected = list(expected_point[['start', 'end']].itertuples(index=False))
    observed = list(observed_point[['start', 'end']].itertuples(index=False))
    expected_return = (None, 4, 4, 2)
    returned = _overlap_segment(expected, observed)
    np.testing.assert_array_equal(np.array(returned),
                                  np.array(expected_return))


def test_contextual_confusion_matrix(expected, observed):
    expected_return = (3, 6, 6, 8)
    returned = contextual_confusion_matrix(expected, observed)
    np.testing.assert_array_equal(np.array(returned),
                                  np.array(expected_return))


def test_contextual_confusion_matrix_points(expected_point, observed_point):
    expected_return = (4, 7, 3, 5)
    returned = contextual_confusion_matrix(expected_point, observed_point)
    np.testing.assert_array_equal(np.array(returned),
                                  np.array(expected_return))


def test_contextual_confusion_matrix_overlap(expected, observed):
    expected_return = (None, 1, 1, 5)
    returned = contextual_confusion_matrix(expected, observed, weighted=False)
    np.testing.assert_array_equal(np.array(returned),
                                  np.array(expected_return))


def test_contextual_confusion_matrix_overlap_points(expected_point, observed_point):
    expected_return = (None, 1, 1, 5)
    returned = contextual_confusion_matrix(expected_point, observed_point, weighted=False)
    np.testing.assert_array_equal(np.array(returned),
                                  np.array(expected_return))


def test_contextual_accuracy(expected, observed):
    expected_return = float(11 / 23)
    returned = contextual_accuracy(expected, observed)
    assert returned == expected_return


def test_contextual_accuracy_overlap(expected, observed):
    with pytest.raises(ValueError):
        contextual_accuracy(expected, observed, weighted=False)


def test_contextual_precision(expected, observed):
    expected_return = float(8 / 14)
    returned = contextual_precision(expected, observed)
    assert returned == expected_return


def test_contextual_recall(expected, observed):
    expected_return = float(8 / 14)
    returned = contextual_recall(expected, observed)
    assert returned == expected_return


def test_contextual_f1_score(expected, observed):
    expected_return = float(4 / 7)
    returned = contextual_f1_score(expected, observed)
    assert returned == expected_return


def test_contextual_f1_score_nan():
    expected = pd.DataFrame({"start": [2], "end": [5]})
    observed = pd.DataFrame({"start": [6], "end": [8]})
    returned = contextual_f1_score(expected, observed)
    assert np.isnan(returned)



import os

import numpy as np
import pandas as pd
from mlblocks.discovery import load_pipeline

from orion.core import Orion


class TestOrion:

    @classmethod
    def setup_class(cls):
        cls.clean = pd.DataFrame({
            'timestamp': list(range(100)),
            'value': [1] * 100,
        })
        cls.anomalous = pd.DataFrame({
            'timestamp': list(range(100, 200)),
            'value': [1] * 45 + [10] * 10 + [1] * 45
        })
        cls.events = pd.DataFrame([
            {'start': 145, 'end': 155, 'severity': 9.0}
        ], columns=['start', 'end', 'severity'])

        cls.all_data = pd.concat((cls.clean, cls.anomalous))
        cls.all_events = pd.DataFrame([
            {'start': 145, 'end': 155, 'severity': 4.275}
        ], columns=['start', 'end', 'severity'])

    def setup(self):
        self.orion = Orion('dummy')

    def test_fit(self):
        self.orion.fit(self.clean)

    def test_detect(self):
        self.orion.fit(self.clean)

        events = self.orion.detect(self.anomalous)

        pd.testing.assert_frame_equal(self.events, events)

    def test_detect_no_visualization(self):
        self.orion.fit(self.clean)

        events, visualization = self.orion.detect(self.anomalous, visualization=True)

        pd.testing.assert_frame_equal(self.events, events)

        assert visualization == {}

    def test_detect_visualization(self):
        pipeline = load_pipeline('dummy')
        pipeline['outputs'] = {
            'visualization': [
                {
                    'name': 'y_hat',
                    'variable': 'orion.primitives.estimators.MeanEstimator#1.y'
                }
            ]
        }
        orion = Orion(pipeline)
        orion.fit(self.clean)

        events, visualization = orion.detect(self.anomalous, visualization=True)

        pd.testing.assert_frame_equal(self.events, events)

        assert isinstance(visualization, dict)
        assert 'y_hat' in visualization
        y_hat = visualization['y_hat']
        np.testing.assert_array_equal(y_hat, np.ones(len(self.anomalous)))

    def test_fit_detect(self):
        events = self.orion.fit_detect(self.all_data)

        pd.testing.assert_frame_equal(self.all_events, events)

    def test_save_load(self, tmpdir):
        path = os.path.join(tmpdir, 'some/path.pkl')
        self.orion.save(path)

        new_orion = Orion.load(path)
        assert new_orion == self.orion

    def test_evaluate(self):
        self.orion.fit(self.clean)
        scores = self.orion.evaluate(data=self.anomalous, ground_truth=self.events)

        expected = pd.Series({
            'accuracy': 1.0,
            'f1': 1.0,
            'recall': 1.0,
            'precision': 1.0,
        })
        pd.testing.assert_series_equal(expected, scores)

    def test_evaluate_fit(self):
        scores = self.orion.evaluate(
            data=self.all_data,
            ground_truth=self.all_events,
            fit=True,
        )

        expected = pd.Series({
            'accuracy': 1.0,
            'f1': 1.0,
            'recall': 1.0,
            'precision': 1.0,
        })
        pd.testing.assert_series_equal(expected, scores)

    def test_evaluate_train_data(self):
        scores = self.orion.evaluate(
            data=self.anomalous,
            ground_truth=self.events,
            fit=True,
            train_data=self.clean
        )

        expected = pd.Series({
            'accuracy': 1.0,
            'f1': 1.0,
            'recall': 1.0,
            'precision': 1.0,
        })
        pd.testing.assert_series_equal(expected, scores)



# -*- coding: utf-8 -*-

__author__ = """MIT Data To AI Lab"""
__email__ = 'dailabmit@gmail.com'
__version__ = '0.3.1.dev0'

import os

from orion.core import Orion
from orion.functional import detect_anomalies, evaluate_pipeline, fit_pipeline

_BASE_PATH = os.path.abspath(os.path.dirname(__file__))
MLBLOCKS_PRIMITIVES = os.path.join(_BASE_PATH, 'primitives', 'jsons')
MLBLOCKS_PIPELINES = tuple(
    dirname
    for dirname, _, _ in os.walk(os.path.join(_BASE_PATH, 'pipelines'))
)

__all__ = (
    'Orion',
    'detect_anomalies',
    'evaluate_pipeline',
    'fit_pipeline'
)



# -*- coding: utf-8 -*-

import logging
import math
import tempfile
from functools import partial

import keras
import numpy as np
import pandas as pd
from keras import backend as K
from keras.layers import Input
from keras.layers.merge import _Merge
from keras.models import Model
from mlprimitives.adapters.keras import build_layer
from mlprimitives.utils import import_object
from scipy import stats

from orion.primitives.timeseries_errors import reconstruction_errors

LOGGER = logging.getLogger(__name__)


class RandomWeightedAverage(_Merge):
    def _merge_function(self, inputs):
        """
        Args:
            inputs[0] x     original input
            inputs[1] x_    predicted input
        """
        alpha = K.random_uniform((64, 1, 1))
        return (alpha * inputs[0]) + ((1 - alpha) * inputs[1])


class TadGAN(object):
    """TadGAN model for time series reconstruction.

    Args:
        shape (tuple):
            Tuple denoting the shape of an input sample.
        encoder_input_shape (tuple):
            Shape of encoder input.
        generator_input_shape (tuple):
            Shape of generator input.
        critic_x_input_shape (tuple):
            Shape of critic_x input.
        critic_z_input_shape (tuple):
            Shape of critic_z input.
        layers_encoder (list):
            List containing layers of encoder.
        layers_generator (list):
            List containing layers of generator.
        layers_critic_x (list):
            List containing layers of critic_x.
        layers_critic_z (list):
            List containing layers of critic_z.
        optimizer (str):
            String denoting the keras optimizer.
        learning_rate (float):
            Optional. Float denoting the learning rate of the optimizer. Default 0.005.
        epochs (int):
            Optional. Integer denoting the number of epochs. Default 2000.
        latent_dim (int):
            Optional. Integer denoting dimension of latent space. Default 20.
        batch_size (int):
            Integer denoting the batch size. Default 64.
        iterations_critic (int):
            Optional. Integer denoting the number of critic training steps per one
            Generator/Encoder training step. Default 5.
        hyperparameters (dictionary):
            Optional. Dictionary containing any additional inputs.
    """

    def __getstate__(self):
        networks = ['critic_x', 'critic_z', 'encoder', 'generator']
        modules = ['optimizer', 'critic_x_model', 'critic_z_model', 'encoder_generator_model']

        state = self.__dict__.copy()

        for module in modules:
            del state[module]

        for network in networks:
            with tempfile.NamedTemporaryFile(suffix='.hdf5', delete=False) as fd:
                keras.models.save_model(state.pop(network), fd.name, overwrite=True)

                state[network + '_str'] = fd.read()

        return state

    def __setstate__(self, state):
        networks = ['critic_x', 'critic_z', 'encoder', 'generator']
        for network in networks:
            with tempfile.NamedTemporaryFile(suffix='.hdf5', delete=False) as fd:
                fd.write(state.pop(network + '_str'))
                fd.flush()

                state[network] = keras.models.load_model(fd.name)

        self.__dict__ = state

    def _build_model(self, hyperparameters, layers, input_shape):
        x = Input(shape=input_shape)
        model = keras.models.Sequential()

        for layer in layers:
            built_layer = build_layer(layer, hyperparameters)
            model.add(built_layer)

        return Model(x, model(x))

    def _setdefault(self, kwargs, key, value):
        if key in kwargs:
            return

        if key in self.hyperparameters and self.hyperparameters[key] is None:
            kwargs[key] = value

    def _wasserstein_loss(self, y_true, y_pred):
        return K.mean(y_true * y_pred)

    def _gradient_penalty_loss(self, y_true, y_pred, averaged_samples):
        gradients = K.gradients(y_pred, averaged_samples)[0]
        gradients_sqr = K.square(gradients)
        gradients_sqr_sum = K.sum(gradients_sqr, axis=np.arange(1, len(gradients_sqr.shape)))
        gradient_l2_norm = K.sqrt(gradients_sqr_sum)
        gradient_penalty = K.square(1 - gradient_l2_norm)
        return K.mean(gradient_penalty)

    def __init__(self, layers_encoder, layers_generator, layers_critic_x, layers_critic_z,
                 optimizer, input_shape=(100, 1), target_shape=(100, 1), latent_dim=20,
                 learning_rate=0.0005, epochs=2000, batch_size=64, iterations_critic=5,
                 **hyperparameters):

        self.shape = input_shape
        self.epochs = epochs
        self.batch_size = batch_size
        self.latent_dim = latent_dim
        self.latent_shape = (latent_dim, 1)
        self.target_shape = target_shape
        self.iterations_critic = iterations_critic

        self.layers_encoder, self.layers_generator = layers_encoder, layers_generator
        self.layers_critic_x, self.layers_critic_z = layers_critic_x, layers_critic_z

        self.optimizer = import_object(optimizer)(learning_rate)

        self.hyperparameters = hyperparameters
        self._fitted = False

    def _augment_hyperparameters(self, X, y, kwargs):
        shape = np.asarray(X)[0].shape
        length = shape[0]
        target_shape = np.asarray(y)[0].shape

        # to infer the shape
        self.shape = self.shape or shape
        self.target_shape = self.target_shape or target_shape

        self._setdefault(kwargs, 'generator_reshape_dim', length // 2)
        self._setdefault(kwargs, 'generator_reshape_shape', (length // 2, 1))
        self._setdefault(kwargs, 'encoder_reshape_shape', self.latent_shape)

        return kwargs

    def _set_shapes(self):
        self.encoder_input_shape = self.shape
        self.generator_input_shape = self.latent_shape
        self.critic_x_input_shape = self.target_shape
        self.critic_z_input_shape = self.latent_shape

    def _build_tadgan(self, **kwargs):
        hyperparameters = self.hyperparameters.copy()
        hyperparameters.update(kwargs)

        self.encoder = self._build_model(hyperparameters, self.layers_encoder,
                                         self.encoder_input_shape)
        self.generator = self._build_model(hyperparameters, self.layers_generator,
                                           self.generator_input_shape)
        self.critic_x = self._build_model(hyperparameters, self.layers_critic_x,
                                          self.critic_x_input_shape)
        self.critic_z = self._build_model(hyperparameters, self.layers_critic_z,
                                          self.critic_z_input_shape)

        self.generator.trainable = False
        self.encoder.trainable = False

        x = Input(shape=self.shape)
        y = Input(shape=self.target_shape)
        z = Input(shape=(self.latent_dim, 1))

        x_ = self.generator(z)
        z_ = self.encoder(x)
        fake_x = self.critic_x(x_)
        valid_x = self.critic_x(y)

        interpolated_x = RandomWeightedAverage()([y, x_])
        validity_interpolated_x = self.critic_x(interpolated_x)
        partial_gp_loss_x = partial(self._gradient_penalty_loss, averaged_samples=interpolated_x)
        partial_gp_loss_x.__name__ = 'gradient_penalty'
        self.critic_x_model = Model(inputs=[y, z], outputs=[valid_x, fake_x,
                                                            validity_interpolated_x])
        self.critic_x_model.compile(loss=[self._wasserstein_loss, self._wasserstein_loss,
                                          partial_gp_loss_x], optimizer=self.optimizer,
                                    loss_weights=[1, 1, 10])

        fake_z = self.critic_z(z_)
        valid_z = self.critic_z(z)
        interpolated_z = RandomWeightedAverage()([z, z_])
        validity_interpolated_z = self.critic_z(interpolated_z)
        partial_gp_loss_z = partial(self._gradient_penalty_loss, averaged_samples=interpolated_z)
        partial_gp_loss_z.__name__ = 'gradient_penalty'
        self.critic_z_model = Model(inputs=[x, z], outputs=[valid_z, fake_z,
                                                            validity_interpolated_z])
        self.critic_z_model.compile(loss=[self._wasserstein_loss, self._wasserstein_loss,
                                          partial_gp_loss_z], optimizer=self.optimizer,
                                    loss_weights=[1, 1, 10])

        self.critic_x.trainable = False
        self.critic_z.trainable = False
        self.generator.trainable = True
        self.encoder.trainable = True

        z_gen = Input(shape=(self.latent_dim, 1))
        x_gen_ = self.generator(z_gen)
        x_gen = Input(shape=self.shape)
        z_gen_ = self.encoder(x_gen)
        x_gen_rec = self.generator(z_gen_)
        fake_gen_x = self.critic_x(x_gen_)
        fake_gen_z = self.critic_z(z_gen_)

        self.encoder_generator_model = Model([x_gen, z_gen], [fake_gen_x, fake_gen_z, x_gen_rec])
        self.encoder_generator_model.compile(loss=[self._wasserstein_loss, self._wasserstein_loss,
                                                   'mse'], optimizer=self.optimizer,
                                             loss_weights=[1, 1, 10])

    def _fit(self, X, target):
        fake = np.ones((self.batch_size, 1))
        valid = -np.ones((self.batch_size, 1))
        delta = np.ones((self.batch_size, 1))

        indices = np.arange(X.shape[0])
        for epoch in range(1, self.epochs + 1):
            np.random.shuffle(indices)
            X_ = X[indices]
            y_ = target[indices]

            epoch_g_loss = []
            epoch_cx_loss = []
            epoch_cz_loss = []

            minibatches_size = self.batch_size * self.iterations_critic
            num_minibatches = int(X_.shape[0] // minibatches_size)

            for i in range(num_minibatches):
                minibatch = X_[i * minibatches_size: (i + 1) * minibatches_size]
                y_minibatch = y_[i * minibatches_size: (i + 1) * minibatches_size]

                for j in range(self.iterations_critic):
                    x = minibatch[j * self.batch_size: (j + 1) * self.batch_size]
                    y = y_minibatch[j * self.batch_size: (j + 1) * self.batch_size]
                    z = np.random.normal(size=(self.batch_size, self.latent_dim, 1))
                    epoch_cx_loss.append(
                        self.critic_x_model.train_on_batch([y, z], [valid, fake, delta]))
                    epoch_cz_loss.append(
                        self.critic_z_model.train_on_batch([x, z], [valid, fake, delta]))

                epoch_g_loss.append(
                    self.encoder_generator_model.train_on_batch([x, z], [valid, valid, y]))

            cx_loss = np.mean(np.array(epoch_cx_loss), axis=0)
            cz_loss = np.mean(np.array(epoch_cz_loss), axis=0)
            g_loss = np.mean(np.array(epoch_g_loss), axis=0)
            print('Epoch: {}/{}, [Dx loss: {}] [Dz loss: {}] [G loss: {}]'.format(
                epoch, self.epochs, cx_loss, cz_loss, g_loss))

    def fit(self, X, y=None, **kwargs):
        """Fit the TadGAN.

        Args:
            X (ndarray):
                N-dimensional array containing the input training sequences for the model.
            y (ndarray):
                N-dimensional array containing the target sequences we want to reconstruct.
        """
        if y is None:
            y = X.copy()  # reconstruct the same input

        if not self._fitted:
            self._augment_hyperparameters(X, y, kwargs)
            self._set_shapes()
            self._build_tadgan(**kwargs)

        self._fit(X, y)
        self._fitted = True

    def predict(self, X, y=None):
        """Predict values using the initialized object.

        Args:
            X (ndarray):
                N-dimensional array containing the input sequences for the model.
            y (ndarray):
                N-dimensional array containing the target sequences we want to reconstruct.

        Returns:
            ndarray:
                N-dimensional array containing the reconstructions for each input sequence.
            ndarray:
                N-dimensional array containing the critic scores for each input sequence.
        """
        if y is None:
            y = X.copy()  # reconstruct the same input

        z_ = self.encoder.predict(X)
        y_hat = self.generator.predict(z_)
        critic = self.critic_x.predict(y)

        return y_hat, critic


def _compute_critic_score(critics, smooth_window):
    """Compute an array of anomaly scores.

    Args:
        critics (ndarray):
            Critic values.
        smooth_window (int):
            Smooth window that will be applied to compute smooth errors.

    Returns:
        ndarray:
            Array of anomaly scores.
    """
    critics = np.asarray(critics)
    l_quantile = np.quantile(critics, 0.25)
    u_quantile = np.quantile(critics, 0.75)
    in_range = np.logical_and(critics >= l_quantile, critics <= u_quantile)
    critic_mean = np.mean(critics[in_range])
    critic_std = np.std(critics)

    z_scores = np.absolute((np.asarray(critics) - critic_mean) / critic_std) + 1
    z_scores = pd.Series(z_scores).rolling(
        smooth_window, center=True, min_periods=smooth_window // 2).mean().values

    return z_scores


def score_anomalies(y, y_hat, critic, index, score_window=10, critic_smooth_window=None,
                    error_smooth_window=None, smooth=True, rec_error_type="point", comb="mult",
                    lambda_rec=0.5):
    """Compute an array of anomaly scores.

    Anomaly scores are calculated using a combination of reconstruction error and critic score.

    Args:
        y (ndarray):
            Ground truth.
        y_hat (ndarray):
            Predicted values. Each timestamp has multiple predictions.
        index (ndarray):
            time index for each y (start position of the window)
        critic (ndarray):
            Critic score. Each timestamp has multiple critic scores.
        score_window (int):
            Optional. Size of the window over which the scores are calculated.
            If not given, 10 is used.
        critic_smooth_window (int):
            Optional. Size of window over which smoothing is applied to critic.
            If not given, 200 is used.
        error_smooth_window (int):
            Optional. Size of window over which smoothing is applied to error.
            If not given, 200 is used.
        smooth (bool):
            Optional. Indicates whether errors should be smoothed.
            If not given, `True` is used.
        rec_error_type (str):
            Optional. The method to compute reconstruction error. Can be one of
            `["point", "area", "dtw"]`. If not given, 'point' is used.
        comb (str):
            Optional. How to combine critic and reconstruction error. Can be one
            of `["mult", "sum", "rec"]`. If not given, 'mult' is used.
        lambda_rec (float):
            Optional. Used if `comb="sum"` as a lambda weighted sum to combine
            scores. If not given, 0.5 is used.

    Returns:
        ndarray:
            Array of anomaly scores.
    """

    critic_smooth_window = critic_smooth_window or math.trunc(y.shape[0] * 0.01)
    error_smooth_window = error_smooth_window or math.trunc(y.shape[0] * 0.01)

    step_size = 1  # expected to be 1

    true_index = index  # no offset

    true = [item[0] for item in y.reshape((y.shape[0], -1))]

    for item in y[-1][1:]:
        true.extend(item)

    critic_extended = list()
    for c in critic:
        critic_extended.extend(np.repeat(c, y_hat.shape[1]).tolist())

    critic_extended = np.asarray(critic_extended).reshape((-1, y_hat.shape[1]))

    critic_kde_max = []
    pred_length = y_hat.shape[1]
    num_errors = y_hat.shape[1] + step_size * (y_hat.shape[0] - 1)

    for i in range(num_errors):
        critic_intermediate = []

        for j in range(max(0, i - num_errors + pred_length), min(i + 1, pred_length)):
            critic_intermediate.append(critic_extended[i - j, j])

        if len(critic_intermediate) > 1:
            discr_intermediate = np.asarray(critic_intermediate)
            try:
                critic_kde_max.append(discr_intermediate[np.argmax(
                    stats.gaussian_kde(discr_intermediate)(critic_intermediate))])
            except np.linalg.LinAlgError:
                critic_kde_max.append(np.median(discr_intermediate))
        else:
            critic_kde_max.append(np.median(np.asarray(critic_intermediate)))

    # Compute critic scores
    critic_scores = _compute_critic_score(critic_kde_max, critic_smooth_window)

    # Compute reconstruction scores
    rec_scores, predictions = reconstruction_errors(
        y, y_hat, step_size, score_window, error_smooth_window, smooth, rec_error_type)

    rec_scores = stats.zscore(rec_scores)
    rec_scores = np.clip(rec_scores, a_min=0, a_max=None) + 1

    # Combine the two scores
    if comb == "mult":
        final_scores = np.multiply(critic_scores, rec_scores)

    elif comb == "sum":
        final_scores = (1 - lambda_rec) * (critic_scores - 1) + lambda_rec * (rec_scores - 1)

    elif comb == "rec":
        final_scores = rec_scores

    else:
        raise ValueError(
            'Unknown combination specified {}, use "mult", "sum", or "rec" instead.'.format(comb))

    true = [[t] for t in true]
    return final_scores, true_index, true, predictions
